/*
 * $Header: /home/CVS/mysqlbench/mysqlbench.c,v 1.9 2004/12/26 17:03:01 takeshi Exp $
 *
 * pgbench: a simple TPC-B like benchmark program for PostgreSQL
 * written by Tatsuo Ishii
 *
 * Copyright (c) 2000  Tatsuo Ishii
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of the
 * author not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission. The author makes no representations about the
 * suitability of this software for any purpose.  It is provided "as
 * is" without express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h> /* RAND_MAX */
#include <mysql.h>

#include <errno.h>

#include <errmsg.h>
#include <mysqld_error.h>

#include <pthread.h>

#ifdef WIN32
#include "win32.h"
#else
#include <sys/time.h>
#include <unistd.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

/* for getrlimit */
#include <sys/resource.h>
#endif   /* WIN32 */


#define TRUE (1)


#define MAX_CHECK_LOOP 2^27

/********************************************************************
 * some configurable parameters */

#define MAXCLIENTS 1024			/* max number of clients allowed */

int			nclients = 1;		/* default number of simulated clients */
int			nxacts = 10;		/* default number of transactions per
								 * clients */

/*
 * scaling factor. for example, tps = 10 will make 1000000 tuples of
 * accounts table.
 */

int			tps = 1;

/*
 * end of configurable parameters
 *********************************************************************/

#define nbranches	1
#define ntellers	10
#define naccounts	100000

int			remains;			/* number of remained clients */
pthread_mutex_t mymutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond_con = PTHREAD_COND_INITIALIZER ;
pthread_cond_t cond_q   = PTHREAD_COND_INITIALIZER ;

int go_flag = 0;
int go_connect = 0;

int	   is_connect = 0;			/* establish connection  for each transactoin */

char	   *mysqlhost = "";
int	   mysqlport = 0;
char	   *login = NULL;
char	   *pwd = NULL;
char	   *dbName = "test";

#define ENGINE_INNODB "innodb"
#define ENGINE_MYISAM "myisam"
char *engine = ENGINE_MYISAM;


typedef struct
{
  MYSQL	                *con;			/* connection handle to DB */
  MYSQL                   mysql;
  pthread_t               ptid;
  int                     n;
  enum {INIT, WAKEUP, CONNECT, Q0, Q1, Q2, Q3, Q4, Q5, Q6}	state;			/* state No. */

  int			cnt;			/* xacts count */
  int			ecnt;			/* error count */
  int			listen;			/* none 0 indicates that an async query has been sent ; for PG */
  int			aid;			/* account id for this transaction */
  int			bid;			/* branch id for this transaction */
  int			tid;			/* teller id for this transaction */
  int			delta;
  int			abalance;
}	CState;



int	debug = 0;		/* debug flag */

/*===================================*/
static void usage()
{
  fprintf(stderr, "usage:\n  mysqlbench [-h hostname] [-p port] [-c nclients] [-t ntransactions] [-s scaling_factor] [-n] [-v] [-S] [-U login] [-P password] [-d] [dbname]\n");
  fprintf(stderr, "(initialize mode):  mysqlbench -i [-h hostname] [-p port] [-s scaling_factor] [-U login] [-P password] [-d] [-E innodb|myisam] [dbname]\n");
}


/*===================================*/
/* random number generator */
static int getrand(int min, int max)
{
  return (min + (int) (max * 1.0 * rand() / (RAND_MAX + 1.0)));
}

/*===================================*/
/* set up a connection to the backend */
MYSQL *doConnect(MYSQL *mysql)
{
  MYSQL *con;
  int i;

  mysql_init(mysql);

  mysql_options(mysql, MYSQL_READ_DEFAULT_GROUP, "mysqlbench");

  for (i=0; i < 3; i++) {
    con = mysql_real_connect(mysql, mysqlhost, login, pwd, dbName, mysqlport, NULL, 0);

    if (con) { return con; }
  }

  if (mysql_errno(mysql) == CR_CONNECTION_ERROR)
  {
    fprintf(stderr,
	    "Check that mysqld is running and that the socket: exists!\n"
	    );
  }
  else if (mysql_errno(mysql) == CR_CONN_HOST_ERROR ||
	   mysql_errno(mysql) == CR_UNKNOWN_HOST)
  {
    fprintf(stderr,"Check that mysqld is running on %s", mysqlhost);
    fprintf(stderr," and that the port is %d.\n",
	    mysqlport ? mysqlport : mysql_port);
    fprintf(stderr,"You can check this by doing 'telnet %s %d'\n",
	    mysqlhost, mysqlport ? mysqlport : mysql_port);
  }
  else
  {    
    fprintf(stderr,"Got error: %s\n", mysql_error(mysql));
  }

  return NULL;
}

/*===================================*/
/* throw away response from backend */
static void discard_response(CState * state)
{
  int   res;

  do
  {
    res = mysql_errno(state->con);
  } while (res);
}


/*===================================*/
/* check to see if the SQL result was good */
static int check(CState *state, int r)
{
  CState	   *st = state;

  if (r != 0)
  {
    fprintf(stderr, "Client %d aborted in state %d: %s\n", st->n, st->state, mysql_error(st->con));
    pthread_mutex_lock(&mymutex);
    remains--;				/* I've aborted */
    pthread_mutex_unlock(&mymutex);
    mysql_close(st->con);
    st->con = NULL;
    return (-1);
  }
  return (0);					/* OK */
}



/*===================================*/

/* process a transaction */
static void doOne(CState *state)
{
  char		sql[512];
  MYSQL_RES   *res;
  CState	   *st;
  int n, v;

  st = state;

  st->state = WAKEUP;

  /* ------------------------------------- */

  /* wait */
  pthread_mutex_lock(&mymutex);
  do {
    pthread_cond_wait(&cond_con, &mymutex);
    //fprintf(stderr, "  wait0 : %d\n", st->n);
  } while (0 == go_connect);
  pthread_mutex_unlock(&mymutex);

  if (NULL == st->con) {
    st->con = doConnect( &(st->mysql) );
  }


  if (NULL == st->con) {
    st->ecnt++;
    return;
  }


  st->state = CONNECT;


  /* wait transaction start flag. */
  pthread_mutex_lock(&mymutex);
  do {
    //fprintf(stderr, "  wait 1 : %d\n", st->n);
    pthread_cond_wait(&cond_q, &mymutex); // cant work clenly
  } while (0 == go_flag);
  pthread_mutex_unlock(&mymutex);

  if (debug) { fprintf(stderr, "%d go\n", st->n); }


  /* ------------------------------------- */
  for (n=0; n < nxacts; n++)
  {
    st->state = Q0;
    strcpy(sql, "BEGIN");
    st->aid = getrand(1, naccounts * tps);
    st->bid = getrand(1, nbranches * tps);
    st->tid = getrand(1, ntellers * tps);
    st->delta = getrand(1, 1000);

    v = mysql_query(st->con, sql);
    if (check(st, v) != 0)	{ st->ecnt++; }

    if (NULL == st->con) { break; }


    st->state = Q1;
    sprintf(sql, "UPDATE accounts SET abalance = abalance + %d WHERE aid = %d", st->delta, st->aid);
    v = mysql_query(st->con, sql);
    if (check(st, v) != 0)	{ st->ecnt++; }

    if (NULL == st->con) { break; }


    st->state = Q2;
    sprintf(sql, "SELECT abalance FROM accounts WHERE aid = %d", st->aid);
    v = mysql_query(st->con, sql);
    if (check(st, v) != 0)
    { st->ecnt++; }
    else
    {
      res = mysql_store_result(st->con);
      mysql_free_result(res);
    }

    if (NULL == st->con) { break; }


    st->state = Q3;
    sprintf(sql, "UPDATE tellers SET tbalance = tbalance + %d WHERE tid = %d", st->delta, st->tid);
    v = mysql_query(st->con, sql);
    if (check(st, v) != 0)	{ st->ecnt++; }

    if (NULL == st->con) { break; }


    st->state = Q4;
    sprintf(sql, "UPDATE branches SET bbalance = bbalance + %d WHERE bid = %d", st->delta, st->bid);
    v = mysql_query(st->con, sql);
    if (check(st, v) != 0)	{ st->ecnt++; }

    if (NULL == st->con) { break; }


    st->state = Q5;
    sprintf(sql, "INSERT INTO history (tid,bid,aid,delta,mtime) VALUES (%d,%d,%d,%d, now() )", st->tid, st->bid, st->aid, st->delta);
    v = mysql_query(st->con, sql);
    if (check(st, v) != 0)	{ st->ecnt++; }

    if (NULL == st->con) { break; }


    st->state = Q6;
    strcpy(sql, "COMMIT");
    v = mysql_query(st->con, sql);
    if (check(st, v) != 0)	{ st->ecnt++; }

    if (NULL == st->con) { break; }


    if (st->con) {
      st->cnt++;
    }

  }

  if (debug) { fprintf(stderr, "%d end\n", st->n); }

}




/*===================================*/
/* discard connections */
static void disconnect_all(CState * state)
{
  int			i;

  for (i = 0; i < nclients; i++)
  {
    if (state[i].con)
    {
      mysql_close(state[i].con);
    }
  }
}



/*===================================*/
/* create tables and setup data */
static void init(MYSQL *mysql)
{
  MYSQL	   *con;
  int        res;

  static char *DDLs[] = {
		"DROP TABLE IF EXISTS branches",
		"CREATE TABLE branches (bid int PRIMARY KEY, bbalance int, filler char(88))",
		"DROP TABLE IF EXISTS tellers",
		"CREATE TABLE tellers (tid int PRIMARY KEY, bid int, tbalance int, filler char(84))",
		"DROP TABLE IF EXISTS accounts",
		"CREATE TABLE accounts (aid int PRIMARY KEY, bid int, abalance int, filler char(84))",
		"DROP TABLE IF EXISTS history",
		"CREATE TABLE history (tid int, bid int, aid int, delta int, mtime timestamp, filler char(22))"
  };
  char		sql[512];

  int			i;

  if ((con = doConnect(mysql)) == NULL)
  {
    fprintf(stderr, "can't connect. exit.\n");
    exit(1);
  }

  for (i = 0; i < (sizeof(DDLs) / sizeof(char *)); i++)
  {
    if (strncmp(DDLs[i], "CREATE", 6) == 0)
    {
      sprintf(sql, "%s TYPE=%s", DDLs[i], engine);
    }
    else
    {
      sprintf(sql, "%s", DDLs[i]);
    }

    res = mysql_query(con, sql);
    if (res != 0)
    {
      fprintf(stderr, "%s\n%s\n", mysql_error(con), sql);
      exit(1);
    }
  }

  res = mysql_query(con, "BEGIN");
  if (res != 0)
  {
    fprintf(stderr, "%s\n", mysql_error(con));
    exit(1);
  }

  for (i = 0; i < nbranches * tps; i++)
  {
    sprintf(sql, "INSERT INTO branches (bid, bbalance) VALUES (%d, 0)", i + 1);
    res = mysql_query(con, sql);
    if (res != 0)
    {
      fprintf(stderr, "%s\n", mysql_error(con));
      exit(1);
    }
  }

  for (i = 0; i < ntellers * tps; i++)
  {
    sprintf(sql, "INSERT INTO tellers (tid, bid, tbalance) VALUES (%d, %d, 0)", i + 1, i / ntellers + 1);
    res = mysql_query(con, sql);
    if (res != 0)
    {
      fprintf(stderr, "%s\n", mysql_error(con));
      exit(1);
    }
  }


  /*
   * occupy accounts table with some data
   */
  fprintf(stderr, "creating tables...\n");
  for (i = 0; i < naccounts * tps; i++)
  {
    int			j = i + 1;

    sprintf(sql, "INSERT INTO accounts (aid, bid, abalance) VALUES (%d, %d, %d)", j, j / naccounts, 0);
    mysql_query(con, sql);

  }


  res = mysql_query(con, "COMMIT");
  if (res != 0)
  {
    fprintf(stderr, "%s\n", mysql_error(con));
    exit(1);
  }


  /* vacuum */
  fprintf(stderr, "OPTIMIZE TABLE...");
  res = mysql_query(con, "OPTIMIZE TABLE branches, tellers, accounts, history");
  if (res != 0)
  {
    fprintf(stderr, "%s\n", mysql_error(con));
    //    exit(1);
  }
  fprintf(stderr, "done.\n");

  mysql_close(con);
}



/*===================================*/
/* print out results */
static void printResults(
			 int ttype, CState * state,
			 struct timeval * tv1, struct timeval * tv2,
			 struct timeval * tv3)
{
  double		t1, t2;
  int			i;
  int			normal_xacts = 0;

  for (i = 0; i < nclients; i++)
  {
    normal_xacts += state[i].cnt;
  }

  t1 = (tv3->tv_sec - tv1->tv_sec) * 1000000.0 + (tv3->tv_usec - tv1->tv_usec);
  t1 = normal_xacts * 1000000.0 / t1;

  t2 = (tv3->tv_sec - tv2->tv_sec) * 1000000.0 + (tv3->tv_usec - tv2->tv_usec);
  t2 = normal_xacts * 1000000.0 / t2;

  printf("transaction type                    . . . : %s\n", ttype == 0 ? "TPC-B (sort of)" : "SELECT only");
  printf("scaling factor                      . . . : %d\n", tps);
  printf("number of clients                     . . : %d\n", nclients);
  printf("number of transactions per client         : %d\n", nxacts);
  printf("number of transactions actually processed : %d/%d\n", normal_xacts, nxacts * nclients);
  printf("tps ( include connections establishing)   : %f\n", t1);
  printf("tps ( exclude connections establishing)   : %f\n", t2);
}





/*===================================*/

int main(int argc, char **argv)
{
  pthread_t	ptid;
  extern char *optarg;
  extern int	optind, opterr,	optopt;
  int			c;
  int			is_init_mode = 0;		/* initialize mode? */
  int			is_no_vacuum = 0;		/* no vacuum at all before testing? */
  int			is_full_vacuum = 0;		/* do full vacuum before testing? */
  int			ttype = 0;		/* transaction type. 0: TPC-B, 1: SELECT only */

  static CState state[MAXCLIENTS];	/* clients status */

  struct timeval tv1;			/* start up time */
  struct timeval tv2;			/* after establishing all connections to
					 * the backend */
  struct timeval tv3;			/* end time */

  int			i, j, k;

  fd_set		input_mask;
  int			nsocks;			/* return from select(2) */
  int			maxsock;		/* max socket number to be waited */

#ifndef __CYGWIN__
  struct rlimit rlim;
#endif

  MYSQL	   *con, mysql0;
  MYSQL_RES   *result;
  int res;

  while ((c = getopt(argc, argv, "ih:nvp:dc:t:s:U:P:CSE:")) != -1)
  {
    switch (c)
    {
      case 'i':
	is_init_mode++;
	break;
      case 'h':
	mysqlhost = optarg;
	break;
      case 'n':
	is_no_vacuum++;
	break;
      case 'v':
	is_full_vacuum++;
	break;
      case 'p':
	if (optarg != NULL) {
	  mysqlport = atoi(optarg);
	}
	break;
      case 'd':
	debug++;
	break;
      case 'S':
	ttype = 1;
	break;
      case 'c':
	nclients = atoi(optarg);
	if (nclients <= 0 || nclients > MAXCLIENTS)
	{
	  fprintf(stderr, "wrong number of clients: %d\n", nclients);
	  exit(1);
	}
#ifndef __CYGWIN__
#ifdef RLIMIT_NOFILE			/* most platform uses RLIMIT_NOFILE */
	if (getrlimit(RLIMIT_NOFILE, &rlim) == -1)
	{
#else							/* but BSD doesn't ... */
	  if (getrlimit(RLIMIT_OFILE, &rlim) == -1)
	  {
#endif   /* HAVE_RLIMIT_NOFILE */
	    fprintf(stderr, "getrlimit failed. reason: %s\n", strerror(errno));
	    exit(1);
	  }
	  if (rlim.rlim_cur <= (nclients + 2))
	  {
	    fprintf(stderr, "You need at least %d open files resource but you are only allowed to use %ld.\n", nclients + 2, (long) rlim.rlim_cur);
	    fprintf(stderr, "Use limit/ulimt to increase the limit before using mysqlbench.\n");
	    exit(1);
	  }
#endif   /* #ifndef __CYGWIN__ */
	  break;
/*
	  case 'C':
	    is_connect = 1;
	    break;
*/
	  case 's':
	    tps = atoi(optarg);
	    if (tps <= 0)
	    {
	      fprintf(stderr, "wrong scaling factor: %d\n", tps);
	      exit(1);
	    }
	    break;
	  case 't':
	    nxacts = atoi(optarg);
	    if (nxacts <= 0)
	    {
	      fprintf(stderr, "wrong number of transactions: %d\n", nxacts);
	      exit(1);
	    }
	    break;
	  case 'U':
	    login = optarg;
	    break;
	  case 'P':
	    pwd = optarg;
	    break;
	  case 'E':
	    engine = optarg;
	    break;
	  default:
	    usage();
	    exit(1);
	    break;
	}
    }

    if (argc > optind)
    {
      dbName = argv[optind];
    }

    if (is_init_mode)
    {
      init(&mysql0);
      exit(0);
    }



    remains = nclients;



    if (debug)
    {
      printf("mysqlhost: %s mysqlport: %s nclients: %d nxacts: %d dbName: %s\n",
	     mysqlhost, mysqlport, nclients, nxacts, dbName);
    }


    /* ---------------------------------------------------- */
    /* opening connection... */
    con = doConnect(&mysql0);
    if (con == NULL)
    {
      fprintf(stderr, "cant connect. at pre stage\n");
      exit(1);
    }


    /*
     * get the scaling factor that should be same as count(*) from
     * branches...
     */
    res = mysql_query(con, "SELECT count(bid) FROM branches");

    if (res != 0)
    {
      fprintf(stderr, "%s\n", mysql_error(con));
      exit(1);
    }

    result = mysql_store_result(con);
    tps = atoi( (mysql_fetch_row(result))[0] );
    if (tps < 0)
    {
      fprintf(stderr, "count(*) FROM branches invalid (%d)\n", tps);
      exit(1);
    }
    mysql_free_result(result);


    if (!is_no_vacuum)
    {
      fprintf(stderr, "starting vacuum...");
      res = mysql_query(con, "OPTIMIZE TABLE branches");
      if (res != 0)
      {
	fprintf(stderr, "OPTIMIZE TABLE branches: %s\n", mysql_error(con));
//			exit(1);
      }

      res = mysql_query(con, "OPTIMIZE TABLE tellers");
      if (res != 0)
      {
	fprintf(stderr, "OPTIMIZE TABLE tellers: %s\n", mysql_error(con));
//			exit(1);
      }

      res = mysql_query(con, "DELETE FROM history");
      if (res != 0)
      {
	fprintf(stderr, "DELETE FROM history: %s\n", mysql_error(con));
//			exit(1);
      }

      res = mysql_query(con, "OPTIMIZE TABLE history");
      if (res != 0)
      {
	fprintf(stderr, "OPTIMIZE TABLE history: %s\n", mysql_error(con));
//			exit(1);
      }

      fprintf(stderr, "end.\n");

      if (is_full_vacuum)
      {
	fprintf(stderr, "starting full vacuum...");
	res = mysql_query(con, "OPTIMIZE TABLE accounts");
	if (res != 0)
	{
	  fprintf(stderr, "OPTIMIZE TABLE accounts: %s\n", mysql_error(con));
//				exit(1);
	}
	fprintf(stderr, "end.\n");
      }
    }
    mysql_close(con);

    /* ---------------------------------------------------- */
    /* ---------------------------------------------------- */

    for (i = 0; i < nclients; i++)
    {
      state[i].con = NULL;
      state[i].n = i;
      state[i].state = INIT;
    }


    /* create thread */
    for (i = 0; i < nclients; i++)
    {
      pthread_create(&(state[i].ptid), NULL, (void*) &doOne, &(state[i]) );
    }
    sleep(1);

    /* all thread OK? */
    for (j=0; j < MAX_CHECK_LOOP; j++)
    {
      k = 0;
      for (i = 0; i < nclients; i++)
      {
	if (state[i].state != INIT  &&  state[i].ptid)
	{
	  k++;
	}
      }

      if (k >= nclients) { break; }
    }

    if (debug) { fprintf(stderr, "  make check num j = %d\n", j); }

    if (k < nclients)
    {
      fprintf(stderr, "can not make thread: %d\n", nclients - k);
      exit(1);
    }


    /* ---------------------------------------------------- */
    /* set random seed */
    gettimeofday(&tv1, NULL);
    srand((uint) tv1.tv_usec);

    /* get start up time */
    gettimeofday(&tv1, NULL);


    pthread_mutex_lock(&mymutex);
    go_connect = 1;
    pthread_cond_broadcast(&cond_con);
    pthread_mutex_unlock(&mymutex);




    /* ---------------------------------------------------- */
    /* all connections is OK? */
    for (j=0; j < MAX_CHECK_LOOP; j++)
    {
      k = 0;
      for (i = 0; i < nclients; i++)
      {
	if (state[i].con != NULL  &&  CONNECT == state[i].state)
	{
	  k++;
	}
      }

      if (k >= nclients) { break; }
    }

    if (debug) { fprintf(stderr, "  con check num j = %d\n", j); }

    if (k < nclients)
    {
      fprintf(stderr, "can not make connection: %d\n", nclients - k);
      exit(1);
    }

    /* time after connections set up */
    gettimeofday(&tv2, NULL);


    fprintf(stderr, "all connection OK\n");

    /* ---------------------------------------------------- */

    fprintf(stderr, "go\n");
    /* go test */
    pthread_mutex_lock(&mymutex);
    go_flag = 1;
    pthread_cond_broadcast(&cond_q);
    pthread_mutex_unlock(&mymutex);

    for (j=0; j < MAX_CHECK_LOOP; j++)
    {
      k = 0;
      for (i = 0; i < nclients; i++)
      {
	if (state[i].state != CONNECT)
	{
	  k++;
	}
	else {
	  pthread_cond_broadcast(&cond_q);
	}
      }

      if (k >= nclients) { break; }
    }

    if (debug) { fprintf(stderr, "  go check num j = %d\n", j); }

    if (k < nclients) {
      fprintf(stderr, "can not start: %d\n", nclients - k);
      exit(1);
    }


    /* ---------------------------------------------------- */

    //Join all threads 
    for (i = 0; i < nclients; i++)
    {
      pthread_join(state[i].ptid, NULL);
    }
		

    fprintf(stderr, " end.\n\n");


    if (remains > 0)
    {
      //      disconnect_all(state);

      /* get end time */
      gettimeofday(&tv3, NULL);
      printResults(ttype, state, &tv1, &tv2, &tv3);
      exit(0);
    }

}
